/* Micro Java: graphical UI for VM (mah 99-5-10)
   ===============================
*/
package MJ.VM;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.text.*;
import java.io.*;

// receiver for callbacks of VM
class GraphIO implements inOutVM {
  JTextArea in, out;
  JLabel state;
  JTextField file;
  int inPos;

  GraphIO (JTextArea i, JTextArea o, JLabel st, JTextField f) {
    in = i; out = o; state = st; file = f;
  }

  public boolean reset () {
    String name = file.getText();
    inPos = 0;
    if (name.length()==0) return true;

    try {
      Reader r = new FileReader(file.getText());  // fetch automatic input from file
      in.read(r, null);
    } catch (IOException ex) {
      out.append("\nError opening input: "+ex.getMessage());
      return false;
    }

    return true;
  }

  public void print (byte b) {
    out.append(""+(char)b);
  }

  public byte read () throws IOException {
    String s;
    try {
      s = in.getDocument().getText(inPos++, 1);
    } catch (BadLocationException ex) {
      throw new IOException("end of input");
    }
    return (byte)s.charAt(0);
  }

  public void stopped (boolean err) {
    if (err) state.setText("execution aborted");
    else state.setText("ready");
  }
}

class UI {
  static int uis;
  VM vm;  // VM of this window
  JFrame frame; // the window
  JTextArea in, out;
  GraphIO graphio; // interace for callbacks of vm
  JTextField classFile, inputFile;
  JLabel state;
  JCheckBox debug;

  UI (VM vm, String className) { // init UI for given parameters
    this.vm = vm;
    uis++;  // count open windows
    initFrame(className);
    vm.inout = graphio = new GraphIO(in, out, state, inputFile);
  }

  void initFrame (String className) {
    in = new JTextArea("", 5, 60);
    out = new JTextArea("", 10, 60);
    out.setFont(new Font("Monospaced", Font.PLAIN, 12));

    Component inView = new JScrollPane(in,
                        JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
                        JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    Component outView = new JScrollPane(out,
                        JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
                        JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    JSplitPane io = new JSplitPane(JSplitPane.VERTICAL_SPLIT, false, inView, outView);

    JPanel ioPanel = new JPanel();
    ioPanel.setLayout(new BorderLayout());
    ioPanel.add(io, BorderLayout.CENTER);
    ioPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createLineBorder(Color.black), "Input and Output"));

    JPanel panel = new JPanel();
    panel.setLayout(new BorderLayout());
    panel.add(ioPanel, BorderLayout.CENTER);
    panel.add(getControls(className), BorderLayout.NORTH);

    frame = new JFrame("MicroJava Interpreter");
    frame.addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {close(e);}
    });
    frame.getContentPane().add("Center", panel);
    frame.pack();
    frame.setVisible(true);
  }

  JPanel getControls (String className) {
     JButton start = new JButton("Start");
    start.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {start();}
    });
    JButton stop = new JButton("Stop");
    stop.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {stop();}
    });
    classFile = new JTextField(vm.className, 20);
    classFile.setText(className);
    classFile.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {reload();}
    });
    JButton reloadClass = new JButton("Reload");
    reloadClass.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {reload();}
    });
    JButton clear = new JButton("Clear");
    clear.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {clear();}
    });
    JButton clone = new JButton("Clone");
    clone.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {copy();}
    });
    debug = new JCheckBox("debug", vm.debug);
    debug.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {vm.debug = debug.isSelected();}
    });
    state = new JLabel("not initialized");
    if (vm.ready) state.setText("ready");
    inputFile = new JTextField("", 20);

    JPanel textPanel = new JPanel(new GridLayout(0, 1));
    textPanel.add(classFile); textPanel.add(inputFile);
    JPanel labelPanel = new JPanel(new GridLayout(0, 1));
    labelPanel.add(new JLabel("Class file: "));
    labelPanel.add(new JLabel("Input file:"));
    JPanel inPanel = new JPanel(new BorderLayout());
    inPanel.add(textPanel, BorderLayout.EAST);
    inPanel.add(labelPanel, BorderLayout.WEST);

    JPanel buttons = new JPanel(new GridLayout(2, 3));
    buttons.add(start); buttons.add(stop); buttons.add(clone);
    buttons.add(clear); buttons.add(reloadClass); buttons.add(debug);

    JPanel pane = new JPanel();
    pane.setLayout(new BorderLayout());
    pane.add(inPanel, BorderLayout.EAST);
    pane.add(buttons, BorderLayout.WEST);
    JPanel line = new JPanel(); line.add(state);
    pane.add(line, BorderLayout.SOUTH);

    return pane;
  }

  void close (WindowEvent e) { // window has been closed
    vm.stop = true; // indicate VM to stop execution
    uis--;  // dec number of open windows
    if (uis==0) System.exit(0); // no more windows => exit JavaVM
  }

  void start() { //start interpretation
    int i;
    if (graphio.reset() && !vm.running && vm.ready) {
      state.setText("running");
      vm.reset();
      vm.interpret();
    }
  }

  void stop() {
    vm.stop = true;
  }

  void clear() {
   out.setText("");
  }

  void copy() {   // clone window (same classFile)
    VM vm = new VM(null);
    try {
      vm.load(classFile.getText());
    } catch (Exception ex) {}
    new UI(vm, classFile.getText());
  }

  void reload() { // reloads the class file
    stop(); clear();
    try {
      while (vm.running) Thread.sleep(1); // wait for VM to stop execution
      vm.load(classFile.getText());
      state.setText("ready");
    } catch (Exception ex) {
      state.setText("not initialized");
      out.setText("Error reading class file '"+classFile.getText()+"'\n"+ex.getMessage());
    }
  }
}












