package MJ.VM;

import java.io.*;
import MJ.CodeGen.Code;

public class VM implements Runnable {
  byte code[];	// code array
  int data[];		// global data
  int heap[];		// dynamic heap
  int dataSize, startPC;
  Stack pStack, eStack; // program stack and expression stack
  int fp, sp, free, pc; // frame ptr, stack ptr, next free block, program counter
  boolean stop, error, running, ready;	// should stop, had error, is running, is ready
  inOutVM inout; // callbacks for I/O
  String className; // name of current interpreted class
  boolean debug = false; // debug output on or off
  Thread thread;  // interpretation thread

  public static final int  // instruction codes
    load        =  1,
    load_0      =  2,
		load_1      =  3,
		load_2      =  4,
		load_3      =  5,
    store       =  6,
    store_0     =  7,
		store_1     =  8,
		store_2     =  9,
		store_3     = 10,
    getstatic   = 11,
    putstatic   = 12,
    getfield    = 13,
    putfield    = 14,
    const_0     = 15,
		const_1     = 16,
		const_2     = 17,
		const_3     = 18,
		const_4     = 19,
		const_5     = 20,
    const_m1    = 21,
    const_      = 22,
    add         = 23,
    sub         = 24,
    mul         = 25,
    div         = 26,
    rem         = 27,
    neg         = 28,
    shl         = 29,
    shr         = 30,
    inc         = 31,
    new_        = 32,
    newarray    = 33,
    aload       = 34,
    astore      = 35,
    baload      = 36,
    bastore     = 37,
    arraylength = 38,
    pop         = 39,
    dup         = 40,
    dup2        = 41,
    jmp         = 42,
    jeq         = 43, jcond = 43,
		jne         = 44,
		jlt         = 45,
		jle         = 46,
		jgt         = 47,
		jge         = 48,
    call        = 49,
    return_     = 50,
    enter       = 51,
    exit        = 52,
    read        = 53,
    print       = 54,
    bread       = 55,
    bprint      = 56,
    trap        = 57;
  public static final int  // compare operators
    eq = 0,
    ne = 1,
    lt = 2,
    le = 3,
    gt = 4,
    ge = 5;

  VM(inOutVM io) { inout = io; }

  void interpret() {
    thread = new Thread(this);
    thread.start();
  }

  void reset() {
    heap = new int[100000]; // fixed sized heap
    data = new int[dataSize]; // global data as specified in classfile
    pStack = new Stack (400); // program stack with startup size
    eStack = new Stack (30);  // expression stack with startup size
    fp = 0;
    free = 1; // start first free block at 1. no block should start at 0
    pc = startPC;
    error = stop = running = false;
  }

  void load (String name) throws IOException, FormatException {
    int codeSize;
    byte sig[] = new byte[2];

    ready = false;
    DataInputStream in =	new DataInputStream(
															new BufferedInputStream(
																	new FileInputStream(name)));
    in.read(sig, 0, 2);
    if (sig[0] != 'M' || sig[1] != 'J')
			throw new FormatException("wrong marker");
    codeSize = in.readInt();
    if (codeSize <= 0) throw new FormatException("codeSize <= 0");
    dataSize = in.readInt();
    if (dataSize < 0) throw new FormatException("dataSize < 0");
    startPC = in.readInt();
    if (startPC < 0 || startPC >= codeSize)
			throw new FormatException("startPC out of bound");
    code = new byte[codeSize];
    in.read(code, 0, codeSize);
    className = name;
    ready = true;
  }

  final int alloc(int size) { // alloc heap block of size bytes
    int adr = free;
    free += ((size+3)>>2); // skips to next free adr (>>2 to convert byte to word)
    return adr;
  }

  final int readInt () throws IOException { // read int from specified inout
    int prev, b, val = 0;

    prev = ' '; b = inout.read();
    while (b<'0' || b>'9') {
      prev = b; b = inout.read();
    }
    while (b>='0' && b<='9') {
      val = 10*val+b-'0';
      b = inout.read();
    }
    if (prev == '-') val = -val;
    return val;
  }

  final int pop() { // shortcut to pop from eStack
    return eStack.pop();
  }

  final void push (int val) { // shortcut to push on eStack
    eStack.push(val);
  }

  final byte getCodeByte () {
    return code[pc++];
  }

  final short getCodeShort () {
    int val = 0;
    val = setByte(val, 2, code[pc++]);
    val = setByte(val, 3, code[pc++]);
    return (short)val;
  }

  final int getCodeWord () {
    int val = 0;
    val = setByte(val, 0, code[pc++]);
    val = setByte(val, 1, code[pc++]);
    val = setByte(val, 2, code[pc++]);
    val = setByte(val, 3, code[pc++]);
    return val;
  }

  // coding fourbytesin an int. byte 0 is MSB
  final static byte getByte (int val, int off) {
    byte b;
    off = 8*off;
    val = val<<off;
    return (byte)(val>>>24);
  }

  final static int setByte (int val, int off, byte b) { // replace byte in int
    int delta = (3-off)*8;
    int mask = ~(255<<delta); // mask all 1 except on chosen byte
    int by = (((int)b) & 255) << delta;

    return (val & mask) ^ by;
  }

  final void print(String s, boolean always) { // prints debug message
    // always false -> print only when debug flag is on
    if(!debug && !always) return;
    try {
      for (int i=0; i < s.length(); i++)
        inout.print((byte)s.charAt(i));
      inout.print((byte)'\n');
    } catch (IOException ex) {}
  }

  final void printErr(String s) {
    stop = error = true;
    print(s, true);
  }

	public void run () {  // actual interpretation
    byte op, lsize, psize;
    boolean cond;
    int adr, val, val2, off, idx, len, i;
    reset();
    running = true;
    long startTime = System.currentTimeMillis();
    try {
      while (!stop) { // while not ended or aborted
        op = getCodeByte();
        print(""+(pc-1)+": "+ getStr(op), false);  // debug output
        switch((int)op) {
          // load/store local variables
          case load: push(pStack.get(fp + getCodeByte())); break;
          case load_0: case load_1: case load_2: case load_3:
            op -= load_0; // mapping on range 0..3
            push(pStack.get(fp+op));
            break;
          case store: pStack.set(fp + getCodeByte(), pop()); break;
          case store_0: case store_1: case store_2: case store_3:
            op -= store_0; // mapping on range 0..3
            pStack.set(fp + op, pop());
            break;

          // load/store global variables
          case getstatic: push(data[getCodeShort()]); break;
          case putstatic: data[getCodeShort()] = pop(); break;

          // load/store object fields
          case getfield:
            adr = pop();
            if (adr == 0) printErr("null reference used!");
            push(heap[adr + getCodeShort()]);
            break;
          case putfield:
            val = pop();
            adr = pop();
            if (adr == 0) printErr("null reference used!");
            heap[adr + getCodeShort()] = val;
            break;

          // load consts
          case const_0: case const_1: case const_2: case const_3: case const_4: case const_5:
            push(op-const_0); // 25..30 auf 0..5
            break;
          case const_m1: push(-1); break;
          case const_: push(getCodeWord()); break;

          // arithmetic
          case add: push(pop()+pop()); break;
          case sub: push(-pop()+pop()); break;
          case mul: push(pop()*pop()); break;
          case div: val = pop(); push(pop()/val); break;
          case rem: val = pop(); push(pop()%val); break;
          case neg: push(-pop()); break;
          case shl: val = pop(); push(pop()<<val); break;
          case shr: val = pop(); push(pop()>>val); break;
          case inc:
					  off = fp + getCodeByte();
						val = getCodeByte();
            pStack.set(off, pStack.get(off) + val);
            break;

          // Objekterzeugung
          case new_: push(alloc(getCodeShort())); break;
          case newarray:
            val = getCodeByte(); len = pop();
            if (val == 5) adr = alloc(len + 4);
            else adr = alloc(len * 4 + 4);
            heap[adr] = len;
            push(adr + 1);   // skip length field of array
            break;

          // Arrayzugriff
          case aload:
            idx = pop(); adr = pop();
            if (adr == 0) printErr("null reference used!");
            push(heap[adr + idx]);
            break;
          case astore:
            val = pop(); idx = pop(); adr = pop();
            if (adr==0) printErr("null reference used!");
            heap[adr+idx] = val;
            break;
          case baload:
            idx = pop(); adr = pop();
            if (adr == 0) printErr("null reference used!");
            push(getByte(heap[adr + idx/4], idx%4));
            break;
          case bastore:
            val = pop(); idx = pop(); adr = pop();
            if (adr == 0) printErr("null reference used!");
            heap[adr + idx/4] = setByte(heap[adr + idx/4], idx%4, (byte)val);
            break;
          case arraylength:
            adr = pop();
            if (adr == 0) printErr("null reference used!");
            push(heap[adr-1]);
            break;

          // stack manipulation
          case pop: pop(); break;
          case dup: val = pop(); push(val); push(val); break;
          case dup2:
            val = pop(); val2 = pop();
            push(val2); push(val); push(val2); push(val);
            break;

          // jumps
          case jmp: off = getCodeShort(); pc += off-3; break;
          case jeq: case jne: case jlt: case jle: case jgt: case jge:
            off = getCodeShort();
            val2 = pop(); val = pop();
            cond = false;
            switch(op-jcond) {
              case eq: cond = val == val2; break;
              case ne: cond = val != val2; break;
              case lt: cond = val < val2; break;
              case le: cond = val <= val2; break;
              case gt: cond = val > val2; break;
              case ge: cond = val >= val2; break;
            }
            if (cond) pc += off-3;
            break;

          // Methodenaufruf
          case call:
            off = getCodeShort();
            pStack.push(pc);
            pc += off-3;
            break;
          case return_:
            if (pStack.SP() == 0) stop = true;
            else pc = pStack.pop();
            break;
          case enter:
            psize = getCodeByte(); lsize = getCodeByte();
            pStack.push(fp);
            fp = pStack.SP();
            for(i=0; i<lsize; i++) pStack.push(0);
            for(i=psize-1; i>=0; i--) {
              val = pop();
              pStack.set(fp+i, val);
            }
            break;
          case exit:
            pStack.cut(fp);
            fp = pStack.pop();
            break;

          // IO
          case read:
            try {
              val = readInt();
              push(val);
            } catch (IOException ex) {
              print("end of input", true);
              error = stop = true;
            }
            break;
          case print:
            len = pop(); val = pop();
            String s = new Integer(val).toString();
            len = len-s.length();
            try {
              for(i=0; i<len; i++) inout.print((byte)' ');
              for(i=0; i<s.length(); i++) inout.print((byte)s.charAt(i));
            } catch (IOException ex) {
              error = stop = true;
            }
            break;
          case bread:
            try {
              push(inout.read());
            } catch (IOException ex) {
              error = stop = true;
            }
            break;
          case bprint:
            len = pop()-1; val = pop();
            try {
              for(i=0; i<len; i++) inout.print((byte)' ');
              inout.print((byte)val);
            } catch (IOException ex) {
              error = stop = true;
            }
            break;

					case trap:
            print("trap("+getCodeByte()+") Program execution aborted", true);
            error = stop = true;
          default:
            print("wrong opcode "+op, true);
            error = stop = true;
        }
      }
    } catch (Throwable ex) {
      error = true;
      ByteArrayOutputStream bstream = new ByteArrayOutputStream();
      PrintStream stream = new PrintStream(bstream);
      ex.printStackTrace(stream);
      print(bstream.toString(), true);
    } finally {
      running = false;
      thread = null;
      print("Completion took "+(System.currentTimeMillis()-startTime)+" milliseconds", true);
      inout.stopped(error);
    }
  }

	public static String getStr (byte opCode) {
		String s = "";
		switch(opCode) {
			case 1: s = "load"; break;
			case 2: s = "load_0"; break;
			case 3: s = "load_1"; break;
			case 4: s = "load_2"; break;
			case 5: s = "load_3"; break;
			case 6: s = "store"; break;
			case 7: s = "store_0"; break;
			case 8: s = "store_1"; break;
			case 9: s = "store_2"; break;
			case 10: s = "store_3"; break;
			case 11: s = "getstatic"; break;
			case 12: s = "putstatic"; break;
			case 13: s = "getfield"; break;
			case 14: s = "putfield"; break;
			case 15: s = "const_0"; break;
			case 16: s = "const_1"; break;
			case 17: s = "const_2"; break;
			case 18: s = "const_3"; break;
			case 19: s = "const_4"; break;
			case 20: s = "const_5"; break;
			case 21: s = "const_m1"; break;
			case 22: s = "const"; break;
			case 23: s = "add"; break;
			case 24: s = "sub"; break;
			case 25: s = "mul"; break;
			case 26: s = "div"; break;
			case 27: s = "rem"; break;
			case 28: s = "neg"; break;
			case 29: s = "shl"; break;
			case 30: s = "shr"; break;
			case 31: s = "inc"; break;
			case 32: s = "new"; break;
			case 33: s = "newarray"; break;
			case 34: s = "aload"; break;
			case 35: s = "astore"; break;
			case 36: s = "baload"; break;
			case 37: s = "bastore"; break;
			case 38: s = "arraylength"; break;
			case 39: s = "pop"; break;
			case 40: s = "dup"; break;
			case 41: s = "dup2"; break;
			case 42: s = "jmp"; break;
			case 43: s = "jeq"; break;
			case 44: s = "jne"; break;
			case 45: s = "jlt"; break;
			case 46: s = "jle"; break;
			case 47: s = "jgt"; break;
			case 48: s = "jge"; break;
			case 49: s = "call"; break;
			case 50: s = "return"; break;
			case 51: s = "enter"; break;
			case 52: s = "exit"; break;
			case 53: s = "read"; break;
			case 54: s = "print"; break;
			case 55: s = "bread"; break;
			case 56: s = "bprint"; break;
			case 57: s = "trap"; break;
			default: s = "-- error--"; break;
		}
		return s;
	}
}
